import os

POSTGRES_DSN = os.environ.get("PSQL_DSN")
AWS_ENDPOINT = os.environ.get("AWS_ENDPOINT")

SOUNDFONT_URL = "https://github.com/Nospoko/soundfonts/raw/master/GeneralUserGSv1.471/GeneralUserGSv1.471.sf2"
